#!/bin/bash 

# Atomic masses database:
a_mass(){

	i=1
	for argval in "$@"; do
		case $argval in
		H)
			atom[$i]="1837.15d0"
			;;
		H2)
			atom[$i]="3674.3d0"
			;;
		O)
			atom[$i]="29156.96d0"
			;;
		C)
			atom[$i]="21874.66d0"
			;;
		N)
			atom[$i]="25526.06d0"
			;;
		Ti)
			atom[$i]="87256.20d0"
			;;
		M) 
			atom[$i]="207.769444551d0"
			;;
		*)
			echo "The atom you used is not in our database, sorry" >> process.txt
			exit
			;;
		esac

		((i = i + 1))
	done
}

corio_setup(){
	#Set up for the Coriolis constants calculation
	if test -e "Lx"; then rm Lx; fi

	if test -e "Ly"; then rm Ly; fi

	if test -e "Lz"; then rm Lz; fi

	update process.txt "Generation of the eigenvectors matrices Lx, Ly, Lz. They are then moved in the Coriolis folder for the calculation" >> process.txt
	
	mapfile -t files < "eigenvec_vib.dat"

	l=1
	a=0
	while [ $l -le $N_atoms ]; do
		echo ${files[a]} >> "$CORIO_dir/Lx"
		let 'a=a+3'
		let 'l++'
	done
	a=1
	l=1
	while [ $l -le $N_atoms ]; do
		echo ${files[a]} >> "$CORIO_dir/Ly"
		let 'a=a+3'
		let 'l++'
	done
	a=2
	l=1
	while [ $l -le $N_atoms ]; do
		echo ${files[a]} >> "$CORIO_dir/Lz"
		let 'a=a+3'
		let 'l++'
	done

}


print_what(){

	if test -e $2; then rm $2; fi

	case $1 in
	intro)
		echo "
      ----------------------------------------
      -  FDACC 19-01-2023                    -  
      -  Mandelli Giacomo, Luca Corneo       - 
      -  v 0.1                               - 
      -  VPT2+K Anharmonic Coefficients      -
      -  Calculator ready to be coupled to   -
      -  the paradensum and parsctst         -
      -  programs for the SCTST calculations - 
      ---------------------------------------- 
     
" >> $2
	case $program in
		gaussian)
       			 echo "You chose to work with the Gaussian software" >> $2
	     		 ;;
			*)
	       		 exit
	       		 ;;
	esac
	 	;;
	*)
		:
		;;
	esac

}


update(){
	echo -e $2 >> ${lch_dir}/$1
	echo " " >> ${lch_dir}/$1
}

update_2(){
	if [ -e "$lch_dir/$1" ] && ! grep -Fxqi "$2" "$lch_dir/$1" || [ ! -e "$lch_dir/$1" ]; then
		echo $2 >> "${lch_dir}/$1"
		echo " " >> "$lch_dir/$1"
	fi
}

std_err_get(){

	local _errname=$1
	if grep -q -i '#SBATCH -e ' ${lch_dir}/input.dat; then
		local _err_name=$(grep -i '#SBATCH -e ' ${lch_dir}/input.dat | awk '{print $3}')
	elif grep -q -i '#SBATCH --error' ${lch_dir}/input.dat; then
		local _err_name=$(grep -i '#SBATCH --error' ${lch_dir}/input.dat | awk '{print $3}')
	else
	 update process.txt "We have a problem: in your slurm directives you didn't set a name for the standard error file. 
	Try to add something like: #SBATCH --error myJob.err "
		exit
	fi
	 
	eval $_errname="'$_err_name'"
}


std_out_get(){

	local _outname=$1

	if grep -q -i '#SBATCH -o ' ${lch_dir}/input.dat; then
		local _out_name=$(grep -i '#SBATCH -o ' ${lch_dir}/input.dat | awk '{print $3}')
	elif grep -q -i '#SBATCH --output' ${lch_dir}/input.dat; then
		local _out_name=$(grep -i '#SBATCH --output' ${lch_dir}/input.dat | awk '{print $3}')
	else
		update process.txt "You didn't define a std output in your slurm directives, it's fine!!"
	fi

	eval $_outname="'$_out_name'"
}

read_variables(){
	# Name of the input files        
	if [ ${input[0]} = "R" ]; then
		file_name="R.com"
	elif [ ${input[0]} = "TS" ]; then
		file_name="TS.com"
	else
		update process.txt "Molecule name NOT VALID. Please restart and enter the proper name =)"
		exit
	fi

	update_2 process.txt "File name: ${file_name}"

	mol=${input[0]}

	N_atoms=$(grep -E '^ *[a-Z]+ +[-.0-9dDeE]+ +[-.0-9dDeE]+ +[-.0-9dDeE]+ *$' $lch_dir/input.dat | wc -l)
	
	update_2 process.txt "Number of atoms of the molecule: $N_atoms"

	program=${input[1]}

	N_cores=${input[2]}

	hess_time=${input[3]}

	hess_finite=${input[4]}

	ab_opt=${input[5]}

	ab_hess=${input[6]}

	ab_anha=${input[7]}

	bs_opt=${input[8]}

	bs_hess=${input[9]}

	bs_anha=${input[10]}

	launch_command=${input[11]}

	group=${input[12]}

	no_modes=${input[13]}

	fine_ist=$(grep -n ccccccccccccccccc input.dat | tail -1 | sed 's/:.*$//g')
	# Line at which the launching info finishes
	(( fine_ist = fine_ist - 1 ))
	# Line fot the multiplicity 
	(( char_molt = fine_ist + 2 ))
	# Multiplicity chars 
	mult=$(sed -n "${char_molt}p" input.dat)
}


geo_mass_generator(){

	(( char_molt = fine_ist + 2 ))
	(( init_mol = char_molt + 1 ))
	(( fine_mol = init_mol + N_atoms ))

	# Creation of the geometry vector => each position is an atom 
	atoms_mol=$(sed -n "${init_mol},${fine_mol}p" ${lch_dir}/input.dat | awk '{print $1}')
	# Generate a vector containing the atoms and a file called atomi 
	echo $atoms_mol > atomi
	sed -i "s/(Iso.*)//g" atomi
	sed -i "s/ //g" atomi
	
	# Generates the masses data accordingly to the atoms specified. If iso.dat is
	# present, then modify the data submitted to the database to obtain the correct
	# masses.
	if [ -e "$lch_dir/iso.dat" ]; then
		isoatoms=( ${atoms_mol[@]} )
		while read -r line; do
			local addr=$(awk '{print $1}' <<< "$line")
           		local iso=$(awk '{print $2}' <<< "$line")
			(( addr = addr - 1 ))
            		isoatoms[$addr]=$(echo "${isoatoms[$addr]}$iso")
		done < "$lch_dir/iso.dat"
		a_mass ${isoatoms[@]}
	else
		a_mass $atoms_mol
	fi

	# We print in the file mass_vec the mass values for each atom in each cartesian direction 
	# atom[] comes from the function a_mass 
	#Condition if mass_vec exists:
	if test -e "mass_vec"; then
		:
	else	
		for i in "${atom[@]}"; do
			echo ${i} >> mass_vec
			echo ${i} >> mass_vec
			echo ${i} >> mass_vec
		done
	fi
}


eckart_gen(){

	for i in $(eval echo "{1..$N_atoms}"); do
		(( k = i - 1 ))
		echo "${arr[$k]} ${geo_vett[$k]}" >> "$HESS_dir/input_eckart"
	done

	update_2 process.txt "Input file for the Eckart frame setting generated: input_eckart"
}

inertia_calc(){
	
	# Working in HESS_dir !!!
	cd "$HESS_dir"
	"$fdacc_bin/inertia.exe"
	wait
	# We have OUTECKART but also the ROT_CONST

	if ! grep -qi "Equilibrium geometry [angstrom]:" "$lch_dir/fdacc.out"; then
		update_2 "fdacc.out" "Equilibrium geometry [angstrom]:"
		cat "output_eckart" >> "$lch_dir/fdacc.out"
		update_2 "fdacc.out" "Rotational Constants [GHz]:"
		cat "Rot_const" >> "$lch_dir/fdacc.out"
	fi
	
	# Move common files in common (project) folder
	mv output_eckart ../
	mv Rot_const ../
	
	cd ..

	update_2 process.txt "Inertia calculation executed, output geometry in Eckart frame: output_eckart"
}

SPE_timer(){
	
	# Find the remainder for the number of lauches(nodes) required for the calculation
	resto=$(($points_hess % $N_cores))
	# Check if the rem is ZERO: 
	if [ $resto -eq 0 ]; then
		sp=$((($points_hess / $N_cores) - 1))
	else
		# If the rem is not zero we use more launching file 
		sp=$(($points_hess / $N_cores)) # Number of single point calculations to be launched 
	fi

	update_2 process.txt "$points_hess points required for the Hessian calculation. $(($sp + 1)) launching files are required."
}



eigen_extractor(){

	if [ $mol = "R" ]; then
		# The first 6 eigenvalues and eigenvec are eliminated as they are supposed to be rototraslational (smallest eigenvalues)
		sed "1,6d" WW.dat > WW_vib.dat
		if [ "${no_modes[0]}" != "0" ]; then
			comando=$(echo "${nmodes[0]}d")
			p=1
			for ((j = 2;j <= $num_nmodes; j++)); do
				comando+=$(echo ";${nmodes[p]}d")
				((p = p + 1))
			done
			sed -i "${comando}" WW_vib.dat
		fi

		awk '{$1=$1;print}' eigenvec.dat > temp
		if [ "${no_modes[0]}" != "0" ]; then
			cut -d ' ' -f 7- temp > temp2
			comando=$(echo "\$${nmodes[0]}=")
			p=1
			for ((j = 2;j <= $num_nmodes; j++)); do
				comando+=$(echo "\$${nmodes[p]}=")
				((p = p + 1))
			done
			comando+=$(echo "\"\"")
			awk "{${comando}; print \$0}" temp2 > eigenvec_vib.dat
		else
			cut -d ' ' -f 7- temp > eigenvec_vib.dat
		fi
		rm temp
		if test -e "temp2"; then
			rm temp2
		fi
	fi # R or TS

	if [ $mol = "TS" ]; then
		update_2 ${dir}/process.txt "     Note: if you are working with a TS molecule and the imaginary vibrational eigenvalue 
		is smaller (in its absolute value) then the eigenvalues associated with the rotational or translational motion 
		you may have a problem => it's necessary to modify the cut range for the vibrational eigenvalue."
		# Here we should add the possibility to exclude specific HR modes from the vibrational eigvec/val matrices !!
		sed "2,7d" WW.dat > WW_vib.dat
		if [ "${no_modes[0]}" != "0" ]; then
			comando=$(echo "${nmodes[0]}d")
			p=1
			for ((j = 2;j <= $num_nmodes; j++)); do
			comando+=$(echo ";${nmodes[p]}d")
			((p = p + 1))
		done
			sed -i "${comando}" WW_vib.dat
		fi
			awk '{$1=$1;print}' eigenvec.dat > temp
		if [ "${no_modes[0]}" != "0" ]; then
			cut  -d ' ' -f1,8- temp > temp2
			comando=$(echo "\$${nmodes[0]}=")
			p=1
		for ((j = 2;j <= $num_nmodes; j++)); do
			comando+=$(echo "\$${nmodes[p]}=")
			((p = p + 1))
		done
			comando+=$(echo "\"\"")
			awk "{${comando}; print \$0}" temp2 > eigenvec_vib.dat
		else
			cut  -d ' ' -f1,8- temp > eigenvec_vib.dat
		fi
			rm temp
		if test -e "temp2"; then
			rm temp2
		fi
	fi

	# Move common files in common (prj) dir
	mv WW_vib.dat ../
	mv eigenvec_vib.dat ../
}

fdiff_launch_all ()
{
	#
	#	Launches all the xp_* calculations prepared for the fdiff calculation, that must
	#	be situated in the calling directory.
	#
	
	for x in xp_*; do
		${launch_command} "$x" >> "$lch_dir/process.txt"
	done
}

anharm_gen_geos(){
	#
	#	Generate the geometries and input files necessary to compute the derivatives
	#	needed to compute the X matrix.
	#

	# IF you are working with a TS molecule this will change the first eigenvalue in the vector and rewrite it positive in the file!
	if [ $mol = "TS" ]; then 
		sed -i 's/^[ \t]*//' ../WW_vib.dat
		sed -i "1s/^.//g" ../WW_vib.dat
	fi

	# Generate the necessary geometries:  
	${fdacc_bin}/derivative.exe >> "$lch_dir/process.txt"
	update_2 process.txt "In the folder $ANHARM_dir let's proceed with the launch of the calculations"

	# Prepare the launch of the SPE calc !!
	resto=$(($points % $N_cores))
	if [ $resto -eq 0 ]; then
		sp=$((($points / $N_cores) - 1)) # If resto != ZERO we work with an additional node  
	else
		sp=$(($points / $N_cores)) # Number of single point calculations to be launched 
	fi
}


anharm_derivatives(){
	#
	#	Compute the derivatives needed to compute the X matrix.
	#

	# Let's compute the derivatives => run derinotens.f90: 
	${fdacc_bin}/calcolo.exe >> ${lch_dir}/process.txt
	wait
	# Output files: derter, derqua
	update_2 process.txt "Derivatives calculated and stored in the files derter and derqua."
	
	# Enter the folder anharmonic
	update_2 process.txt "We now begin the calculation of the third order and fourth order derivatives contribution to the anharmonic constants" #???

	#rm *.com *.out xp*
	for file in *.com *.out xp*; do
                rm "$file"
        done

}

anharm_compute(){
	#	
	#	Compute the X_matrix and other anharmonicities-related quantities, such
	#	as the anharmonic ZPE and the G0 term, outputting them in the proper files.
	#
	

	# If we are working with the Reactant  
	if [ $mol = "R" ]; then
		G0=($(${fdacc_bin}/anharmonic.exe)) 
		wait
		update_2 fdacc.out "G0 term [cm^-1] (without Coriolis contribution): ${G0[0]}"
	else ## We are working with the TS
		G0=($(${fdacc_bin}/anharmonicTS.exe))
		wait
		update_2 fdacc.out "G0 term [cm^-1] (without Coriolis contribution): ${G0[0]}"
		update_2 fdacc.out "Im contribution to G0 term [cm^-1] (included in G0): ${G0[1]}"
	fi
	# If you wish you can add a file called ORDER.dat in the folder where you run the program and it wil automatically perform a coordinate change in the out.dat file => useful if you want a comparison with gaussian results or if you wish for any particular reason to have a specific normal mode ordering in the matrix output.  
	# Test if there is an ORDER file!
	if test -e ../ORDER.dat; then
		update_2 process.txt "A file ORDER.dat has been found in the main directory. The output anharmonic constants matrix will be rearranged in the desired order"
		${fdacc_bin}/order.exe # Generates a new out.dat file, a new freq file newFREQ.dat and a new Corio file !!
	fi
	update_2 process.txt "The file out.dat contains the anharmonic constants matrix (third and fourth order contribution)"
	update_2 process.txt "If a new ordering is desired, the file newFREQ.dat will contain the normal modes in the new ordered fashion"

	# Compute the sum of coriolis contributions and derivatives ones to the ANHA

	if [ $mol = "R" ]; then
		anhazpe=($(${fdacc_bin}/somma.exe))
		wait
		update_2 fdacc.out "Anharmonic ZPE [cm^-1]: ${anhazpe[0]}"
	else 
		anhazpe=($(${fdacc_bin}/somma.exe -im))
		wait
		update_2 fdacc.out "Anharmonic ZPE [cm^-1]: ${anhazpe[0]}"
	fi
	update_2 process.txt "The total X matrix is stored in the file: X_tot.dat. The additional info about the calculations (G0, anharmonic ZPE ...) are stored in the file: fdacc.out"
}


fdiff_launch_groups (){
	# 1- group; 2- N_launch; 
	
	#	Group = no of launches per group
	#	N_launch = no of Xps

	h=0
	k=$(($1 - 1))
	# Cycle over the total number of packages to be launched 
	# Each package contains K launch files for K nodes
	for j in $(eval echo "{1..$(($2 / $group + $( (( $2 % $group == 0 )) && echo 0 || echo 1 ) ))}"); do
		# j runs over the total number of packages 
		l=1
		echo "Launching calculations from $h to $k" >> ${lch_dir}/process.txt
		echo " " >> ${lch_dir}/process.txt
		# Launching packages from h to k
		for i in $(eval echo "{$h..$k}"); do
			if [ -f xp_${i} ]; then
				if [ $j -eq 1 ]; then # First launching cycle 
					if [ $i -eq $k ]; then # When you launch the last of the group
						if [ ${launch_command} = "qsub" ]; then
							num=$(${launch_command} xp_${i})
						elif [ ${launch_command} = "sbatch" ]; then
							num=$(${launch_command} xp_${i} | awk '{print $4}')
						else 
							update ${lch_dir}/process.txt "sorry, the launching command is not known"
							exit 
						fi

						echo "$num" >> ${lch_dir}/process.txt
					else 
						${launch_command} xp_${i} >> ${lch_dir}/process.txt
					fi
				else
					if [ $i -eq $k ]; then
						if [ ${launch_command} = "qsub" ]; then
							num=$(${launch_command} -W depend=afterok:${num} xp_${i})
						elif [ ${launch_command} = "sbatch" ]; then
							num=$(${launch_command} --dependency=afterok:${num} xp_${i} | awk '{print $4}')
						else
							update ${lch_dir}/process.txt "sorry, the launching command is not known" 
							exit 
						fi
						echo "$num" >> ${lch_dir}/process.txt
					else
						if [ ${launch_command} = "qsub" ]; then
							${launch_command} -W depend=afterok:${num} xp_${i} >> ${lch_dir}/process.txt
						elif [ ${launch_command} = "sbatch" ]; then
							${launch_command} --dependency=afterok:${num} xp_${i} >> ${lch_dir}/process.txt
						else 
							update ${lch_dir}/process.txt "sorry, the launching command is not known"
							exit
						fi
					fi 
				fi
			fi
			(( l = l +1 ))
		done
		# Group launch completed: go with the next ones 
		(( h = h + $1 ))
		(( k = k + $1 ))
	done
}

opt_retriever() {
	#
	#	Extract the informations from the optimization job and ouputs them in process.txt. 
	#	The function assumes that a frequency calculation has been performed along with
	#	the optimization (which is true if the opt has been launched with this script) and 
	#	prints also the harmonic ZPE corrected energies.
	#
	

	# Creation of fdacc.out file
	int_opt_retriever
	
	if [ -e "$lch_dir/fdacc.out" ] && ! grep -iq "Sum of electronic and zero-point Energies:" "$lch_dir/fdacc.out" || [ ! -e "$lch_dir/fdacc.out" ]; then
		update_2 fdacc.out "Sum of electronic and harmonic zero-point Energies [Hartrees]: $zpeEle"
		update_2 fdacc.out "Harmonic zero-point correction [Hartrees]: $zpe"
		update_2 fdacc.out "Electronic energy [Hartrees]: $elec"
		
	fi

}

opt_check(){
	#
	#	Check if the opt job has not termineted yet. If not so, check if there are 
	#	errors in the execution. If any of theese condition is met, EXIT the job.
	#	

	# Check if the job has returned an outpud (so supposedly ended)
	if [ ! -e ${OPT_dir}/${mol}.out ]; then
		echo "Looks like the optimization job has not finished yet."
		exit
	fi

    # Check if there are errors in the output file, calling the interface function
	if [ -s ${def_err_name} ] || int_err_chk ${OPT_dir}/${mol}.out; then 
		echo "Error! Something went wrong in the Gaussian calculation!"
		update process.txt "Error! Something went wrong in the Gaussian calculation!"
		exit
	fi

}

hess_check ()
{
	#
	#	Check if the hess job has not termineted yet. If not so, check if there are 
	#	errors in the execution. If any of these condition is met, EXIT the job.
	#	

	# Check for job termination i.e. if the output is present
	if [ ! -e "$HESS_dir/$mol.out" ]; then
		echo "Looks like the hessian calculation job has not finished yet."
		exit
	fi
	# Check for errors usign interface function.
	if [ -s "$HESS_dir/${def_err_name}" ] || int_err_chk ${HESS_dir}/${mol}.out; then 
		echo "Error! Something went wrong in the Gaussian hessian calculation!"
		update process.txt "Error! Something went wrong in the Gaussian hessian calculation!"
		exit
	fi
}

fdiff_check ()
{
	#
	#	Check if any of the SPE jobs has not termineted yet. If not so, check if there are 
	#	errors in the execution. If any of these condition is met, EXIT the job.
	#
	#	Args:	$1 no of points to check	
	#

	### To check: geo_eq.out, all N.out which are outputs of geoN.com

	local mess1="Looks like the fdiff SPEs job have not finished yet.\nIf the calculation are not running, your workload manager may have cancelled the jobs due to time limit."
	local mess2="An error has been found in some Gaussian SPE calculations for the fdiff computing."

	if [ ! -e "geo_eq.out" ]; then
		echo -e  "$mess1"
		exit
	elif int_err_chk "geo_eq.out"; then
		echo -e  "$mess2"
		exit	
	fi
		
	for i in $(seq 1 $1); do
		if [ ! -e "$i.out" ]; then
			echo -e  "$mess1"
			exit
		elif int_err_chk "$i.out"; then
			echo -e  "$mess2"
			update process.txt "$mess2"
			exit	
		fi
	done
}

fdiff_retriever(){
	#
	#	Generates the file Energy.dat retrieveing the energies in the correct order,
	#	searching this value accordingly to the level of theory selected.
	#	
	#	Args:	$1 no of point to retrieve

	if test -e Energy.dat; then
		rm Energy.dat
	fi
	
	int_energy_ret geo_eq.out >> Energy.dat
	
	# Creates the output vector
	
	# Read and open the files in the vector 
	for i in $(seq 1 $1); do
 		int_energy_ret "$i.out" >> Energy.dat
	done
	
	# Energy.dat created!!
}


