        program gen_geo 

        use constants

        implicit none 
        !New variables:
        integer :: ncart, nf, points, nat
        real*8 :: delta
        character(len=10) :: multcharge
        ! Necessary input information are the displacement delta for the
        ! finite difference calculations and the number of atoms of the
        ! non-linear molecule!!
        ! The default FF representation used is: 3MR coupling scheme. 

        integer :: i, j, k, l, fra, tra

        real*8 :: deltaQ, lam, trash
        real*8, allocatable :: m(:)
        ! Initial geometry can be represented in the following fashion: (Natom,3)
        ! geov is the vector representation of the matrix and as the
        ! displacement vectors and the new geometries is a 3N
        ! dimensional vector. 

        real*8, allocatable :: geo(:,:)
        real*8, allocatable :: geov(:)
        real*8, allocatable :: geonv(:,:)
        real*8, allocatable :: disp(:,:)
        real*8, allocatable :: cnorm(:,:)
        real*8, allocatable :: freq(:)
        
        !Printing: 
        
        character(len=10) :: file_id 
        character(len=50) :: file_name
        character(len=100) :: atoms
        character(len=100) :: dummy
        character(len=20) :: line1
        character(len=60) :: line2
        integer :: y


        !nf = 3*nat-6
        !fra = nf*(nf-1)/2
        !tra = nf*(nf-1)*(nf-2)/6
        !points = 6*nf+12*((nf*(nf-1))/2)+8*((nf*(nf-1)*(nf-2))/6)
        
        !Read all the new variables:
        open (100,file='../parameters.dat', status="old")
        read(100,*) ncart
        read(100,*) nf
        read(100,*) points
        read(100,*) nat
        read(100,*) delta
        read(100,'(A)') multcharge
        close(100)


        allocate(geo(nat,3))
        allocate(geov(ncart))
        allocate(geonv(ncart,points)) 
        allocate(disp(ncart,nf)) ! Each normal mode has its own
        allocate(cnorm(ncart,nf)) ! Column matrix of the eigenvectors     
        allocate(m(ncart)) ! Mass vector 
        ! The order of the following vector must match the one of the
        ! Cnorm matrix: 
        allocate(freq(nf))
        
        ! Reading step: 
        open(unit=5, file='../eigenvec_vib.dat', status='old')
        do i=1, ncart
         read(5,*) (cnorm(i,j), j=1, nf)
        end do 
        close(5)
        open(unit=5, file='../mass_vec', status='old')
        do i=1, ncart
         read(5,*) m(i)
        end do
        close(5)

        open(unit=5, file='geo', status='old')
        do i=1, nat ! N runs on the atoms 
         read(5,*) (geo(i,j), j=1, 3)
        end do 
        close(5)
        
        open(unit=5, file='../WW_vib.dat', status='old')
       
        do i=1, nf ! Number of the eigenvalues
         read(5,*) freq(i) ! Reading of the eigenvalues 
        end do  
        close(5)
        !Transform the initial equilibrium geometry matrix into a vector
        !of length 3N: 
        k=1
        do i=1, nat
         do j=1, 3
          geov(k) = geo(i,j) 
          k = k+1  
         end do 
        end do  
         
 
        !Let's fill the DISPL matrix: 
        do i=1, nf ! Cicle on the eigenvectors
         do j=1, ncart 
          disp(j,i)=cnorm(j,i)*delta*dsqrt(1/m(j))*&
          &dsqrt(1/(dsqrt(freq(i))))
         end do  
        end do 


        !Conversion of the displacement from Angstrom to AU:        
        do i=1, ncart
         do j=1, nf  
          disp(i,j) = disp(i,j)*autoA  
         end do 
        end do 
        
        ! Generation of the points: 
        ! In the 3, 2, 1 tensors we save the geometry associated with
        ! the displacement along mode j: 

        i = 1
        do j=1, nf
         geonv(:,i) = geov(:) + 3.d0*disp(:,j)
         i=i+1
         geonv(:,i) = geov(:) + 2.d0*disp(:,j) 
         i=i+1
         geonv(:,i) = geov(:) + disp(:,j)
         i=i+1
         geonv(:,i) = geov(:) - disp(:,j)
         i=i+1
         geonv(:,i) = geov(:) - 2.d0*disp(:,j)
         i=i+1
         geonv(:,i) = geov(:) - 3.d0*disp(:,j)
         i=i+1
        end do 

        
        !Generation of the mixed displacements: 
 
        do j=1, nf !Vectors of length = N(N-1)/2
         do k=1, j
          if (j .ne. k) then
          geonv(:,i) = geov(:) + disp(:,j) + disp(:,k)
          i = i + 1
          end if
         end do
        end do


        do j=1, nf
         do k=1, nf 
          if (j .ne. k) then ! N(N-1)
          geonv(:,i) = geov(:) + disp(:,j) - disp(:,k)
          i = i + 1
          end if
         end do
        end do


        do j=1, nf
         do k=1, j
          if (j .ne. k) then ! N(N-1)/2
          geonv(:,i) = geov(:) - disp(:,j) - disp(:,k)
          i = i + 1
          end if
         end do
        end do
        
        
        do j=1, nf ! N(N-1)(N-2)/6
         do k=1, j
          do l=1, k
           if ((j .ne. k) .and. (l .ne. k)) then
           geonv(:,i) = geov(:) + disp(:,j) + disp(:,k) + disp(:,l)
           i = i + 1
           end if
          end do
         end do
        end do
        
        
        do j=1, nf ! N(N-1)(N-2)/2
         do k=1, j
          do l=1, nf 
           if ((j .ne. k) .and. (l .ne. k) .and. (l .ne. j)) then
           geonv(:,i) = geov(:) + disp(:,j) + disp(:,k) - disp(:,l)
           i = i + 1
           end if
          end do
         end do
        end do

        do j=1, nf ! N(N-1)(N-2)/2
         do k=1, nf
          do l=1, k
           if ((j .ne. k) .and. (l .ne. k) .and. (l .ne. j)) then
           geonv(:,i) = geov(:) + disp(:,j) - disp(:,k) - disp(:,l)
           i = i + 1
           end if
          end do
         end do
        end do

        do j=1, nf ! N(N-1)(N-2)/6
         do k=1, j
          do l=1, k
           if ((j .ne. k) .and. (l .ne. k)) then
           geonv(:,i) = geov(:) - disp(:,j) - disp(:,k) - disp(:,l)
           i = i + 1
           end if
          end do
         end do
        end do 

        do j=1, nf ! N(N-1)
         do k=1, nf 
          if (j .ne. k) then
          geonv(:,i) = geov(:) + 3.d0*disp(:,j) + disp(:,k)
          i = i + 1
          end if
         end do
        end do


        do j=1, nf ! N(N-1)
         do k=1, nf 
          if (j .ne. k) then
          geonv(:,i) = geov(:) + 3.d0*disp(:,j) - disp(:,k)
          i = i + 1
          end if
         end do
        end do
        do j=1, nf ! N(N-1)
         do k=1, nf
          if (j .ne. k) then
          geonv(:,i) = geov(:) - 3.d0*disp(:,j) - disp(:,k)
          i = i + 1
          end if
         end do
        end do
        do j=1, nf ! N(N-1)
         do k=1, nf
          if (j .ne. k) then
          geonv(:,i) = geov(:) - 3.d0*disp(:,j) + disp(:,k)
          i = i + 1
          end if
         end do
        end do
        
        
        if ((i-1) .eq. points) then ! Control check 
         write(*,*) "OK"
        end if

       ! Beginning of the printing phase:

        open(unit=5, file='../atomi', status='old') 
        read(5, '(A)') atoms 
        close(5)

        open(unit=5, file='../data.dat', status='old')
        do i=1, 8 
         read(5, '(A)') dummy
        end do
        read(5, '(A)') line1
        read(5, '(A)') line2 
        close(5)
        

        do y=1, points
         write(file_id, '(i0)') y
         file_name = 'geo' // trim(adjustl(file_id)) // '.com'
         open(unit=5, file=trim(file_name), status='replace')

         write(5,*) trim(line1)
         write(5,*) trim(line2)
         write(5,*) " "
         write(5,*) "random"
         write(5,*) " "
         write(5,*) multcharge 
         j=1
         l=1
         do while (j .le. ncart) 
           write(5,*) atoms(l:l), (geonv(k,y), k=j, 2+j)
           j = j+3 
           l = j - 2*l
         end do
         write(5,*) " " 
         close(5)       
        end do 

        endprogram gen_geo
