        module insub

        implicit none

        contains

        SUBROUTINE centermass(nat,X,mass,mtot, Xc)
        integer, intent(in) :: nat
        real(8),intent(in)  :: X(1:3*nat)
        real(8),intent(out) :: Xc(1:3)
        real(8),intent(in)  :: mtot, mass(1:3*nat)
        integer             :: j
        integer, parameter  :: D=3
        Xc = 0.d0
        do j = 1,nat
          Xc(1:D) = Xc(1:D) + X((j-1)*D+1:(j-1)*D+D)*mass(D*j)
        enddo
        Xc = Xc / mtot
        END SUBROUTINE

        SUBROUTINE trframe(q,nvec)
        !==========================================!
        ! translate the coordinates q              !
        ! into another frame defined by 3x3 matrix !
        ! nvec                                     !
        !==========================================!
        real(8),dimension(:),intent(inout)::q
        real(8),dimension(1:3,1:3)::nvec
        ! ::::::::::::::::::::
        integer::dim,natm,i
        real(8),dimension(1:3)::pt

        dim=size(q)
        natm=dim/3

        do i=1,natm
          pt=q(3*i-2:3*i)
          q(3*i-2:3*i)=matmul(transpose(nvec),pt)
        end do

        continue
        END SUBROUTINE trframe

        SUBROUTINE paframe(mass,q,nvec,idia)
        !========================================!
        ! convert the coordinates to the         !
        ! Principle Axis Frame                   !
        !========================================!
        real(8),dimension(:),intent(in)::mass
        real(8),dimension(:),intent(inout)::q
        real(8),dimension(1:3,1:3),intent(out)::nvec
        real(8),dimension(1:3),intent(out)::idia
        ! ::::::::::::::::::::
        real(8),dimension(1:3,1:3)::iten

        integer::info,lwork
        real(8),dimension(1:3)::eigvl !Eigen Vaule
        real(8),dimension(1:8)::work !just temporary space
        lwork=8

        call inertia_tensor(mass,q,iten)
        !nvec=iten
        !call rs('v','u',3,nvec,3,eigvl,work,lwork,info)
        call diagonalizer(3,iten,eigvl,nvec)
        idia=eigvl
        call trframe(q,nvec)

        return
        END SUBROUTINE paframe

        SUBROUTINE inertia_tensor(mass,q,iten)
        !============================================!
        ! calculate the inertia tensor of a molecule !
        ! with mass "mass" and coordinates "q"       !
        !============================================!
        real(8),dimension(:),intent(in)::mass,q
        real(8),dimension(1:3,1:3),intent(out)::iten
        ! ::::::::::::::::::::
        integer::dim,natm,i
        real(8)::m,x,y,z

        dim=size(q,1)
        natm=dim/3

        iten=0.0d0
        do i=1,natm
          m=mass(3*i)
          x=q(3*i-2); y=q(3*i-1); z=q(3*i)
          iten(1,1)=iten(1,1)+m*(y**2+z**2)
          iten(2,2)=iten(2,2)+m*(z**2+x**2)
          iten(3,3)=iten(3,3)+m*(x**2+y**2)
          iten(1,2)=iten(1,2)-m*x*y
          iten(1,3)=iten(1,3)-m*x*z
          iten(2,3)=iten(2,3)-m*y*z
        end do
          iten(2,1)=iten(1,2)
          iten(3,1)=iten(1,3)
          iten(3,2)=iten(2,3)
        return
        END SUBROUTINE inertia_tensor

        end module insub
